import re
import numpy as np
import pandas as pd
from os import listdir
from os.path import isfile, join


class read_dx():

    def __init__(self):

        pass


    def read(self, file, start_code="##TITLE", end_code="##END", from_dir=False, files_dir=None):

        start_code_len = len(start_code)
        end_code_len = len(end_code)

        self.Params = {}  # Empty dict to store params
        self.x = []  # Initialize empty list to store data
        self.y = []  # Initialize empty list to store data
        self.Samples = {}  # Dictionary to nest in multiple samples
        self.Y_matrix = np.array([])
        self.Conc = np.array([])
        sample_count = 0
        start_key = None  # Init none before first appear of title
        end_key = False
        param_name = ""

        if from_dir:
            files=files_dir # A list with all files with its directory
        else:
            files=[file]  # A single file to read

        self.FileName = files

        for file in files: # Can  read single or multiple files
            f = open(file, "r")  # Open the file
            # Loop over the lines

            for line in f:
                line = line.replace("\n", "")  # Delete line break

                if line[0:start_code_len] == start_code:
                    _, start_key = re.findall('##(.+)\s*=\s*(.+)', line)[0]  # get the title of the sample
                    self.Samples[start_key] = {}  # Open a new dictionary for given sample defined by start key
                    self.Samples[start_key]['y'] = []  # Empty list to store Y data
                    self.Samples[start_key]['Conc'] = []  # Empty list t o store concentrations in case of needed

                if line[0:2] == '##':  # It is a parameter, starts with....
                    try:
                        # Extract data  as  ##Param = value, remove spaces from group with \s*
                        param_name, value = re.findall('##(.+)\s*=\s*(.+)', line)[0]
                        # Save it as dictionary
                        self.Samples[start_key][param_name] = value
                    except Exception as ex:
                        param_name = re.findall('##(.+)\s*=', line)[0]
                        # print(line,"\n",ex)
                        pass
                else:
                    pass

                # if it doesn´t start with ## then it spectral data
                # extend data into the list
                # each line is read and splitted using +- separator
                # Only read from position 1 to end, since position 0 is X reference
                # Uses also the latest param to enter to  loop
                if param_name == 'CONCENTRATIONS':
                    try:
                        measure, value = re.findall('<(.+)>\s*,\s*(.+),', line)[0]
                        self.Samples[start_key]['Conc'].append(float(value))
                    except Exception as ex:
                        # print(ex,line)
                        pass

                # print(start_key,param_name,self.Samples[start_key]['XYDATA'])
                if start_key != None and param_name == 'XYDATA':
                    if self.Samples[start_key]['XYDATA'] == '(X++(Y..Y))' and param_name == 'XYDATA':

                        # Regex description:
                        # Starts with 0 or 1 '+', then 0 or 1 '-', then digits with 1 or more repetitions
                        # Then a decimal point (0 or 1) and then digits optional
                        splitted = re.findall('\+?\-?\d{1,16}\.?\d{1,16}', line)
                        if len(splitted) > 0:
                            self.Samples[start_key]['y'].extend(splitted[1:])

                    elif self.Samples[start_key]['XYDATA'] == '(XY..XY)' and param_name == 'XYDATA':  # SpectralEngines
                        # Data is in format  X, Y; X, Y;.
                        # Catches numbers that end with ;  and then remove it using a comprehension list
                        splitted = re.findall('\+?\-?\d{1,16}\.?\d{1,16}\;', line)
                        if len(splitted) > 0:
                            self.Samples[start_key]['y'].extend([val.replace(';', '') for val in splitted])
                    else:
                        pass

                if start_key != None and param_name == 'END':

                    if line[0:end_code_len] == end_code:
                        end_key = True

                        if end_key:

                            # Define the end of file and transform to numeric
                            self.Samples[start_key]['X'] = np.linspace(float(
                                self.Samples[start_key]['FIRSTX']), float(self.Samples[start_key]['LASTX'])
                                , int(self.Samples[start_key]['NPOINTS'])) * float(self.Samples[start_key]['XFACTOR'])
                            self.Samples[start_key]['Y'] = np.array(
                                list(map(float, self.Samples[start_key]['y']))) * float(
                                self.Samples[start_key]['YFACTOR'])

                            # Y array matrix concatenation
                            if self.Y_matrix.shape[0] == 0:
                                self.Y_matrix = self.Samples[start_key]['Y']  # 1st time
                                sample_count += 1
                                self.X_cols = self.Samples[start_key]['X']
                            else:
                                self.Y_matrix = np.vstack(
                                    [self.Y_matrix, self.Samples[start_key]['Y']])  # Append and collect matrix
                                sample_count += 1

                            # Concentration matrix array
                            if self.Conc.shape[0] == 0:
                                self.Conc = np.array(self.Samples[start_key]['Conc'])  # 1st time


                            else:
                                self.Conc = np.vstack(
                                    [self.Conc, self.Samples[start_key]['Conc']])  # Append and collect matrix

                            end_key = False  # toggle off end key

        self.Y_matrix.reshape(sample_count, -1)
        df = pd.DataFrame(data=self.Y_matrix)
        if df.shape[1] != len(self.X_cols):
            df = df.transpose()

        df.columns = self.Samples[start_key]['X']
        df.index = list(self.Samples.keys())
        return df


    def read_from_dir(self, Directory='', ext='.dx',start_code="##TITLE", end_code="##END"):
        # Read all files from directory and create a list, also ensures that the extension of file is correct
        Flist = [Directory+"/"+f for f in listdir(Directory) if isfile(join(Directory, f)) and f.endswith(ext)]

        df=self.read(file=None,start_code=start_code, end_code=end_code, from_dir=True, files_dir=Flist)

        return df





import spc
import pandas as pd
from os import listdir
from os.path import isfile, join

def read_spc(filename):

    f=spc.File(filename) #Read file
    #Extract X & y
    if f.dat_fmt.endswith('-xy'):
        for s in f.sub:
            x=s.x
            y=s.y
    else:
        for s in f.sub:
            x = f.x
            y = s.y

    Spec=pd.Series(y,index=x)

    return Spec

def read_spc_dir(Directory, ext='.spc', orient='Row'):

    #Read all files from directory and create a list, also ensures that the extension of file is correct
    Flist = [f for f in listdir(Directory) if isfile(join(Directory, f)) and f.endswith(ext)]

    SpectraDict={}
    #Read all of them an add them to a dictionary
    for file in Flist:
        Spec=read_spc(Directory + "/" + file)
        SpectraDict[file]=Spec

    #Decide the orientation of dataframe, column-wise or row-wise.
    if orient=='Row':
        SpectraDataFrame=pd.DataFrame(SpectraDict).transpose()
    else:
        SpectraDataFrame = pd.DataFrame(SpectraDict)

    return SpectraDataFrame, SpectraDict



import numpy as np
from scipy.signal import savgol_filter
import pandas as pd
class GeneralTransformer():

    def __init__(self):
        pass


    def fit_transform(self,spc):
        self.fit(spc)
        return self.transform(spc)



class snv(GeneralTransformer):
    def __init__(self):
        '''
        Performs standard normal variate transformation.

        '''
        self.MeanSpectra=None
        self.StdSpectra=None

    def fit(self,spc):
        '''
        Calculates Standard normal variate transformation by:
            (xi - x_mean)/x_std
        :param spc: row wise dataframe of spectra. Each row is a different sample
        :return: Fitted object  with mean and std.
        '''
        self.MeanSpectra=spc.mean(axis=0)
        self.StdSpectra=spc.std(axis=0)
        return None

    def transform(self,spc):
        return (spc -self.MeanSpectra)/self.StdSpectra

class msc(GeneralTransformer):
    '''
    This function implements the multiplicative scatter correction method which attempts to remove physical light scatter
     by accounting for additive and multiplicative effects
     The Multiplicative Scatter Correction (MSC) is a normalization method that attempts to account for additive and
      multiplicative effects by aligning each spectrum (x_ix_i) with an ideal reference one (x_rx_r) as follows:

    x_i = m_i x_r + a_ix_i = m_i x_r + a_i MSC(x_i) = a_i - x_im_iMSC(x_i) = a_i - x_i/m_i
    where a_ia_i and m_im_i are the additive and multiplicative terms respectively.
    '''
    def __init__(self):
        '''
        spc: Input dataframe of spectra
        Conduct multiple scattering correction
        '''
        self.MeanSpectra=None

    def fit(self,spc):
        '''

        :param spc: Row wise input data frame
        :return:
        '''
        self.MeanSpectra= np.array(spc.mean(axis=0))

    def transform(self,spc):

        def transformMSC(xi,MeanSpectra):
            '''
            Fits coeffients of linear regression: (Mean spectra, spectra_i)
            :param xi: Each row or sample of spectra .
            :param MeanSpectra: Mean spectra of all samples.
            :return: Linear regression coefficients
            '''
            m,b= np.polyfit(MeanSpectra,xi,1)
            return (xi-b)*m

        return spc.apply(transformMSC,args=(self.MeanSpectra,),axis=1)


class detrend(GeneralTransformer):
    '''
    Normalizes each row of an input data.frame or matrix by  fitting a second order linear model and returning the fitted residuals.
    The detrend is a row-wise transformation that allows to correct for wavelength-dependent scattering effects (variations in curvilinearity).
    A second-degree polynomial is fit through each spectrum
    '''
    def __init__(self):

        self.MeanSpectra=None

    def fit(self,spc):
        pass

    def fit_transform(self,spc,wave,deg=2):
        '''

        :param spc:  Input  dataframe (row wise) with the Abs of each spectrum
        :param wave: Wavelength array vector
        :return:  Detrended  spectra array or matrix
        '''

        detrend_df= spc.apply(lambda x: x- np.polyval(np.polyfit(wave,x,deg), wave),axis=1)
        return  detrend_df


class derivative(GeneralTransformer):

    def __init__(self):

        pass

    def fit(self,spc):
        pass

    def fit_transform(self,spc, d=2,drop=True):
        '''

        :param spc:  Input  row wise dataframe to calculate derivative
        :param d:  Degree of derivative
        :param drop: Drop nan columns after derivation
        :return: Calculated derivated dataframe.
        '''

        df_deriv=spc.diff(d,axis=1)

        if drop:
            cols_to_drop= list(df_deriv.columns)[0:2]
            df_deriv.drop(columns=cols_to_drop,inplace=True)
        return df_deriv


class sav_gol(GeneralTransformer):

    def __init__(self):
        pass

    def fit(self):
        pass

    def transform(self,spc, window=7,poly=3,deriv=0):
        '''

        :param spc: Input row wise dataframe of spectra
        :param window: The length of the filter window (i.e., the number of coefficients). window_length must be a positive odd integer.
        :param poly:The order of the polynomial used to fit the samples. polyorder must be less than window_length.
        :param deriv: The order of the derivative to compute. This must be a nonnegative integer. The default is 0, which means to filter the data without differentiating.
        :return:
        '''

        self.orig_cols=np.array(spc.columns)
        self.min_wave= self.orig_cols.min()
        self.max_wave= self.orig_cols.max()

        df_sg=pd.DataFrame(savgol_filter(spc,window_length=window,polyorder=poly,axis=1,deriv=deriv))
        self.new_cols= np.linspace(self.min_wave,self.max_wave,len(df_sg.columns))

        #Rename index and columns from dataframe
        df_sg.index=spc.index
        df_sg.columns=self.new_cols

        return df_sg
